@@ -1,7 +1,8 @@
 -- | This module allows to use QuickCheck properties in tasty.
-{-# LANGUAGE CPP, GeneralizedNewtypeDeriving, DeriveDataTypeable #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving, DeriveDataTypeable #-}
 module Test.Tasty.QuickCheck
   ( testProperty
+  , testProperties
   , QuickCheckTests(..)
   , QuickCheckReplay(..)
   , QuickCheckShowReplay(..)
@@ -10,12 +11,14 @@ module Test.Tasty.QuickCheck
   , QuickCheckVerbose(..)
   , module Test.QuickCheck
     -- * Internal
-    -- | This is exposed for testing purposes and not considered as a part
-    -- of the public API.
-    -- You probably shouldn't need it.
+    -- | If you are building a test suite, you don't need these functions.
+    --
+    -- They may be used by other tasty add-on packages (such as tasty-hspec).
   , QC(..)
+  , optionSetToArgs
   ) where
 
+import Test.Tasty ( testGroup )
 import Test.Tasty.Providers
 import Test.Tasty.Options
 import qualified Test.QuickCheck as QC
@@ -36,15 +39,14 @@ import Test.QuickCheck hiding -- for re-export
   )
 
 import Data.Typeable
-import Data.Proxy
 import Data.List
 import Text.Printf
+import Test.QuickCheck.Random (mkQCGen)
+import System.Random (getStdRandom, randomR)
+#if !MIN_VERSION_base(4,8,0)
 import Control.Applicative
-
-#if MIN_VERSION_QuickCheck(2,7,0)
-import Test.QuickCheck.Random (QCGen)
-#else
-import System.Random (StdGen)
+import Data.Monoid
+import Data.Proxy
 #endif
 
 newtype QC = QC QC.Property
@@ -54,16 +56,19 @@ newtype QC = QC QC.Property
 testProperty :: QC.Testable a => TestName -> a -> TestTree
 testProperty name prop = singleTest name $ QC $ QC.property prop
 
+-- | Create a test from a list of QuickCheck properties. To be used
+-- with 'Test.QuickCheck.allProperties'. E.g.
+--
+-- >tests :: TestTree
+-- >tests = testProperties "Foo" $allProperties
+testProperties :: TestName -> [(String, Property)] -> TestTree
+testProperties name = testGroup name . map (uncurry testProperty)
+
 -- | Number of test cases for QuickCheck to generate
 newtype QuickCheckTests = QuickCheckTests Int
   deriving (Num, Ord, Eq, Real, Enum, Integral, Typeable)
 
--- | Replay a previous test using a replay token
-#if MIN_VERSION_QuickCheck(2,7,0)
-newtype QuickCheckReplay = QuickCheckReplay (Maybe (QCGen, Int))
-#else
-newtype QuickCheckReplay = QuickCheckReplay (Maybe (StdGen, Int))
-#endif
+newtype QuickCheckReplay = QuickCheckReplay (Maybe Int)
   deriving (Typeable)
 
 -- | If a test case fails unexpectedly, show the replay token
@@ -90,21 +95,22 @@ instance IsOption QuickCheckTests where
 
 instance IsOption QuickCheckReplay where
   defaultValue = QuickCheckReplay Nothing
-  parseValue v = QuickCheckReplay . Just <$> replay
-    -- Reads a replay token in the form "{size} {seed}"
-    where replay = (,) <$> safeRead (intercalate " " seed) <*> safeRead (concat size)
-          (size, seed) = splitAt 1 $ words v
+  -- Reads a replay int seed
+  parseValue v = QuickCheckReplay . Just <$> safeRead v
   optionName = return "quickcheck-replay"
-  optionHelp = return "Replay token to use for replaying a previous test run"
+  optionHelp = return "Random seed to use for replaying a previous test run (use same --quickcheck-max-size)"
 
 instance IsOption QuickCheckShowReplay where
-  defaultValue = QuickCheckShowReplay True
+  defaultValue = QuickCheckShowReplay False
   parseValue = fmap QuickCheckShowReplay . safeRead
   optionName = return "quickcheck-show-replay"
   optionHelp = return "Show a replay token for replaying tests"
 
+defaultMaxSize :: Int
+defaultMaxSize = QC.maxSize QC.stdArgs
+
 instance IsOption QuickCheckMaxSize where
-  defaultValue = fromIntegral $ QC.maxSize QC.stdArgs
+  defaultValue = fromIntegral defaultMaxSize
   parseValue = fmap QuickCheckMaxSize . safeRead
   optionName = return "quickcheck-max-size"
   optionHelp = return "Size of the biggest test cases quickcheck generates"
@@ -120,7 +126,35 @@ instance IsOption QuickCheckVerbose where
   parseValue = fmap QuickCheckVerbose . safeRead
   optionName = return "quickcheck-verbose"
   optionHelp = return "Show the generated test cases"
-  optionCLParser = flagCLParser Nothing (QuickCheckVerbose True)
+  optionCLParser = mkFlagCLParser mempty (QuickCheckVerbose True)
+
+-- | Convert tasty options into QuickCheck options.
+--
+-- This is a low-level function that was originally added for tasty-hspec
+-- but may be used by others.
+--
+-- @since 0.9.1
+optionSetToArgs :: OptionSet -> IO (Int, QC.Args)
+optionSetToArgs opts = do
+  replaySeed <- case mReplay of
+    Nothing -> getStdRandom (randomR (1,999999))
+    Just seed -> return seed
+
+  let args = QC.stdArgs
+        { QC.chatty          = False
+        , QC.maxSuccess      = nTests
+        , QC.maxSize         = maxSize
+        , QC.replay          = Just (mkQCGen replaySeed, 0)
+        , QC.maxDiscardRatio = maxRatio
+        }
+
+  return (replaySeed, args)
+
+  where
+    QuickCheckTests    nTests   = lookupOption opts
+    QuickCheckReplay   mReplay  = lookupOption opts
+    QuickCheckMaxSize  maxSize  = lookupOption opts
+    QuickCheckMaxRatio maxRatio = lookupOption opts
 
 instance IsTest QC where
   testOptions = return
@@ -132,18 +166,19 @@ instance IsTest QC where
     , Option (Proxy :: Proxy QuickCheckVerbose)
     ]
 
-  run opts (QC prop) yieldProgress = do
+  run opts (QC prop) _yieldProgress = do
+    (replaySeed, args) <- optionSetToArgs opts
+
     let
-      QuickCheckTests      nTests     = lookupOption opts
-      QuickCheckReplay     replay     = lookupOption opts
       QuickCheckShowReplay showReplay = lookupOption opts
-      QuickCheckMaxSize    maxSize    = lookupOption opts
-      QuickCheckMaxRatio   maxRatio   = lookupOption opts
       QuickCheckVerbose    verbose    = lookupOption opts
-      args = QC.stdArgs { QC.chatty = False, QC.maxSuccess = nTests, QC.maxSize = maxSize, QC.replay = replay, QC.maxDiscardRatio = maxRatio}
+      maxSize = QC.maxSize args
       testRunner = if verbose
                      then QC.verboseCheckWithResult
                      else QC.quickCheckWithResult
+      replayMsg = makeReplayMsg replaySeed maxSize
+
+    -- Quickcheck already catches exceptions, no need to do it here.
     r <- testRunner args prop
 
     qcOutput <- formatMessage $ QC.output r
@@ -151,11 +186,12 @@ instance IsTest QC where
           if "\n" `isSuffixOf` qcOutput
             then qcOutput
             else qcOutput ++ "\n"
-
+        testSuccessful = successful r
+        putReplayInDesc = (not testSuccessful) || showReplay
     return $
-      (if successful r then testPassed else testFailed)
+      (if testSuccessful then testPassed else testFailed)
       (qcOutputNl ++
-        (if showReplay then reproduceMsg r else ""))
+        (if putReplayInDesc then replayMsg else ""))
 
 successful :: QC.Result -> Bool
 successful r =
@@ -163,9 +199,9 @@ successful r =
     QC.Success {} -> True
     _ -> False
 
--- | If the result is a failure, produce a message that explains how to
--- reproduce it. If the result is not a failure, return an empty string.
-reproduceMsg :: QC.Result -> String
-reproduceMsg QC.Failure { QC.usedSize = size, QC.usedSeed = seed } =
-  printf "Use --quickcheck-replay '%d %s' to reproduce." size (show seed)
-reproduceMsg _ = ""
+makeReplayMsg :: Int -> Int -> String
+makeReplayMsg seed size = let
+    sizeStr = if (size /= defaultMaxSize)
+                 then printf " --quickcheck-max-size=%d" size
+                 else ""
+  in printf "Use --quickcheck-replay=%d%s to reproduce." seed sizeStr
